#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <epicsAssert.h>
#include <epicsThread.h>

#include <asynDriver.h>
#include <asynOctetSyncIO.h>

#include <imca-base64.h>
#include <imca-string-utils.h>
#include <imca-asyn-utils.h>

#include "dliEpcrLib.h"

static asynStatus writeAndReadDevice(asynUser *anAsynUser,
  const char *whatToWrite, char *whatWasRead, int whatWasReadSize,
  size_t *numberOfBytesRead);
static int generateHttpBasicAuthorizationCookie(char *destination,
  int destinationSize, const char *httpUser, const char *httpPassword);
static int writeHttpGetRequestAndRead(asynUser *anAsynUser,
  const char *httpUser, const char *httpPassword, const char *uri,
  char *whatWasRead, int whatWasReadSize, size_t *numberOfBytesRead);
static short parseOutletState(int model, const char *aString, int outlet);

int switchOutlet(asynUser *anAsynUser, int model, const char *httpUser,
    const char *httpPassword, int outlet, short desiredState) {
  char uri[32];
  char response[HTTP_RESPONSE_SIZE];
  size_t numberOfBytesRead;
  int status;
  const char *anAsynUserPortName;
  asynStatus status2;

  response[0] = '\0';
  numberOfBytesRead = 0;
  if (model == DLI_MODEL_PC8000) {
    sprintf(uri, "/outlet%s?%d", desiredState == OUTLET_ON ? "on" : "off",
      outlet);
  } else if (model == DLI_MODEL_EPCR2 || model == DLI_MODEL_LPC2 ||
      model == DLI_MODEL_LPC9) {
    sprintf(uri, "/outlet?%d=%s", outlet,
      desiredState == OUTLET_ON ? "ON" : "OFF");
  } else {
    assert(0); return 1; /* Programmer error: unknown model */
  }
  status = writeHttpGetRequestAndRead(anAsynUser, httpUser, httpPassword, uri,
    response, HTTP_RESPONSE_SIZE, &numberOfBytesRead);
  if (status == 0) {
    response[numberOfBytesRead] = '\0';
  } else {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:switchOutlet:error: "
      "writeHttpGetRequestAndRead returned %d\n", anAsynUserPortName,
      APPLICATION_NAME, status);
    return 1;
  }

  imcaStripTrailingWhitespace(response);
  if (!imcaStringEndsWithIgnoringCase(response, "</html>")) {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:switchOutlet:error: "
      "HTTP response from device did not end with </html>\n",
      anAsynUserPortName, APPLICATION_NAME);
    return 1;
  }

  return 0;
}

int getOutletStatesFromDevice(asynUser *anAsynUser, int model,
    const char *httpUser, const char *httpPassword, short states[]) {
  char *uri;
  char response[HTTP_RESPONSE_SIZE];
  size_t numberOfBytesRead;
  int status;
  const char *anAsynUserPortName;
  asynStatus status2;

  response[0] = '\0';
  numberOfBytesRead = 0;
  if (model == DLI_MODEL_PC8000) {
    uri = "/index.shtml";
  } else if (model == DLI_MODEL_EPCR2 || model == DLI_MODEL_LPC2 ||
      model == DLI_MODEL_LPC9) {
    uri = "/index.htm";
  } else {
    assert(0); return 1; /* Programmer error: unknown model */
  }
  status = writeHttpGetRequestAndRead(anAsynUser, httpUser, httpPassword, uri,
    response, HTTP_RESPONSE_SIZE - 1, &numberOfBytesRead);
  if (status == 0) {
    response[numberOfBytesRead] = '\0';
  } else {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:getOutletStatesFromDevice:"
      "error: writeHttpGetRequestAndRead returned %d\n", anAsynUserPortName,
      APPLICATION_NAME, status);
    return 1;
  }

  status = parseOutletStates(model, response, states);
  if (status != 0) {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:getOutletStatesFromDevice:"
      "error: parseOutletStates returned %d when attempting to parse "
      "response from device to determine outlet states\n", anAsynUserPortName,
      APPLICATION_NAME, status);
    return 1;
  }

  return 0;
}

int parseOutletStates(int model, const char *response, short states[]) {
  char *matchString;
  int i;
  const char *substring;
  short state;

  /* matchString must match on the desired string, but may also match on an
   * undesired string (e.g. "href=outlet?1=CLC>Cycle</a>" for the EPCR2) as
   * long as parseOutletState will parse only the desired string.
   * parseOutletState may be called to parse an undesired string and will
   * return a failure status; the search will skip the undesired string and
   * continue to the next match. */
  if (model == DLI_MODEL_PC8000) {
    matchString = "<a href=outlet";
  } else if (model == DLI_MODEL_EPCR2 || model == DLI_MODEL_LPC2) {
    matchString = " href=outlet?";
  } else if (model == DLI_MODEL_LPC9) {
    matchString = " href=\"outlet?";
  } else {
    assert(0); return 1; /* Programmer error: unknown model */
  }

  substring = response;
  for (i = 0; i < 8; i++) {
    for (;;) {
      state = -1;
      if (substring == NULL) break;
      substring = strstr(substring, matchString);
      if (substring == NULL) break;
      state = parseOutletState(model, substring, i + 1);
      substring++;
      if (state == -1) continue;
      states[i] = state;
      break;
    }
    if (state == -1) return 1;
  }

  return 0;
}

/*
 * Parses aString and returns the state (0=Off,1=On) of the specified outlet
 * (1-8). Returns -1 if aString does not represent the specified outlet or
 * there was an error parsing aString.
 */
static short parseOutletState(int model, const char *aString, int outlet) {
  int howManyAssigned;
  char invertedStateString[4];
  int parsedOutlet;
  char greaterThanAtEnd[2];
  char *offString;

  invertedStateString[0] = '\0';
  parsedOutlet = -1;

  if (model == DLI_MODEL_PC8000) {
    howManyAssigned = sscanf(aString,
      "<a href=outlet%3[^?]?%1d>Switch %*3[^<]</a%1[>]", invertedStateString,
      &parsedOutlet, greaterThanAtEnd);
  } else if (model == DLI_MODEL_EPCR2 || model == DLI_MODEL_LPC2) {
    howManyAssigned = sscanf(aString,
      " href=outlet?%1d=%3[^>]>Switch %*3[^<]</a%1[>]", &parsedOutlet,
      invertedStateString, greaterThanAtEnd);
  } else if (model == DLI_MODEL_LPC9) {
    howManyAssigned = sscanf(aString,
      " href=\"outlet?%1d=%3[^\"]\">Switch %*3[^<]</a%1[>]", &parsedOutlet,
      invertedStateString, greaterThanAtEnd);
  } else {
    assert(0); return 1; /* Programmer error: unknown model */
  }

  if (howManyAssigned != 3 || parsedOutlet != outlet) return -1;

  if (model == DLI_MODEL_PC8000) {
    offString = "off";
  } else if (model == DLI_MODEL_EPCR2 || model == DLI_MODEL_LPC2 ||
      model == DLI_MODEL_LPC9) {
    offString = "OFF";
  } else {
    assert(0); return 1; /* Programmer error: unknown model */
  }

  return strcmp(invertedStateString, offString) == 0 ? 1 : 0;
}

static int writeHttpGetRequestAndRead(asynUser *anAsynUser,
    const char *httpUser, const char *httpPassword, const char *uri,
    char *whatWasRead, int whatWasReadSize, size_t *numberOfBytesRead) {
  char request[HTTP_REQUEST_SIZE];
  int status;
  char authorizationCookie[ENCODED_HTTP_AUTH_COOKIE_SIZE];
  const char *anAsynUserPortName;
  asynStatus status2;
  asynStatus status3;

  if (strlen(uri) > MAX_URI_LENGTH) {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:"
      "writeHttpGetRequestAndRead:error: URI for GET request is too long "
      "(length=%d,max=%d)\n", anAsynUserPortName, APPLICATION_NAME,
      (int)strlen(uri), MAX_URI_LENGTH);
    return 1;
  }

  status = generateHttpBasicAuthorizationCookie(authorizationCookie,
    ENCODED_HTTP_AUTH_COOKIE_SIZE, httpUser, httpPassword);
  if (status != 0) {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:"
      "writeHttpGetRequestAndRead:error: "
      "generateHttpBasicAuthorizationCookie returned %d\n",
      anAsynUserPortName, APPLICATION_NAME, status);
    return 1;
  }

  sprintf(request,
    "GET %s HTTP/1.0\r\n"
    "Authorization: Basic %s\r\n"
    "\r\n",
    uri, authorizationCookie);
  status2 = writeAndReadDevice(anAsynUser, request, whatWasRead,
    whatWasReadSize, numberOfBytesRead);
  if (status2 != asynSuccess) {
    status3 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status3 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:"
      "writeHttpGetRequestAndRead:error: writeAndReadDevice returned %d\n",
      anAsynUserPortName, APPLICATION_NAME, status2);
    return 1;
  }

  return 0;
}

/* Returns 0 on success, 1 if httpUser and httpPassword pair is too long, and
 * 2 if encoding as Base64 failed. */
static int generateHttpBasicAuthorizationCookie(char *destination,
    int destinationSize, const char *httpUser, const char *httpPassword) {
  int status;
  char unencodedCookie[UNENCODED_HTTP_AUTH_COOKIE_SIZE];

  if (strlen(httpUser) + 1 + strlen(httpPassword) >=
      UNENCODED_HTTP_AUTH_COOKIE_SIZE)
    return 1;
  sprintf(unencodedCookie, "%s:%s", httpUser, httpPassword);
  status = imcaEncodeAsBase64(destination, destinationSize, unencodedCookie,
    strlen(unencodedCookie));
  return status == 0 ? 0 : 2;
}

/*
 * Writes to the device, then reads from it.
 *
 * Has special behavior when reading which is why this function is used rather
 * than using the asynOctetSyncIO#writeRead function (or related functions) or
 * an interface provided in imcalib.
 *
 * The device closes the connection immediately after sending its response;
 * this causes an asynError status to be returned from the read call (with
 * anAsynUser->errorMessage indicating that the connection was closed) even if
 * the read was successful. To handle this, if more than 0 bytes are read, the
 * asynError status is ignored and the read is considered successful. Callers
 * should validate what was read to be sure it is actually OK (i.e. the read
 * really was successful).
 */
static asynStatus writeAndReadDevice(asynUser *anAsynUser,
    const char *whatToWrite, char *whatWasRead, int whatWasReadSize,
    size_t *numberOfBytesRead) {
  size_t numberOfBytesWritten;
  asynStatus status;
  int eomReason;
  const char *anAsynUserPortName;
  asynStatus status2;

  numberOfBytesWritten = 0;
  *numberOfBytesRead = 0;
  if (whatWasReadSize > 0) whatWasRead[0] = '\0';

  status = imcaConnectIpClient(anAsynUser);
  if (status != asynSuccess) {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:writeAndReadDevice:error: "
      "imcaConnectIpClient returned %d (%s)\n", anAsynUserPortName,
      APPLICATION_NAME, status, anAsynUser->errorMessage);
    return status;
  }

  status = pasynOctetSyncIO->flush(anAsynUser);
  if (status != asynSuccess) {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:writeAndReadDevice:error: "
      "pasynOctetSyncIO->flush returned %d\n", anAsynUserPortName,
      APPLICATION_NAME, status);
    imcaDisconnectIpClient(anAsynUser);
    return status;
  }

  status = imcaAsynWriteString(anAsynUser, whatToWrite, DEVICE_WRITE_TIMEOUT);
  if (status != asynSuccess) {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:writeAndReadDevice:error: "
      "imcaAsynWriteString returned %d\n", anAsynUserPortName,
      APPLICATION_NAME, status);
    imcaDisconnectIpClient(anAsynUser);
    return status;
  }

  status = pasynOctetSyncIO->read(anAsynUser, whatWasRead, whatWasReadSize,
    DEVICE_READ_TIMEOUT, numberOfBytesRead, &eomReason);
  if (status == asynSuccess && *numberOfBytesRead > 0) {
    if (whatWasReadSize > 0) whatWasRead[*numberOfBytesRead] = '\0';
  } else if (status == asynError && *numberOfBytesRead > 0) {
    if (whatWasReadSize > 0) whatWasRead[*numberOfBytesRead] = '\0';
    status = asynSuccess;
  } else {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:writeAndReadDevice:error: "
      "pasynOctetSyncIO->read returned %d (numberOfBytesRead=%d,eomReason=%d)"
      "\n", anAsynUserPortName, APPLICATION_NAME, status,
      (int)*numberOfBytesRead, eomReason);
    imcaDisconnectIpClient(anAsynUser);
    return status;
  }

  status = imcaDisconnectIpClient(anAsynUser);
  if (status != asynSuccess) {
    status2 = pasynManager->getPortName(anAsynUser, &anAsynUserPortName);
    if (status2 != asynSuccess) anAsynUserPortName = "<unknown>";
    asynPrint(anAsynUser, ASYN_TRACE_ERROR, "%s:%s:writeAndReadDevice:error: "
      "imcaDisconnectIpClient returned %d\n", anAsynUserPortName,
      APPLICATION_NAME, status);
  }

  return status;
}
